Program BIFTool;
Uses DOS,STMR_STR,ASMUtils,Utils;
{ Author: Eddy L O Jansson
    Mail: srm_dfr@hotmail.com
    home: http://hem2.passagen.se/eddy1/
}

{
  v0.0.61 - fixed unneccesary creation of empty directories
  v0.0.62 - Added ID 0x3F8; '.EFF'

}

CONST
 VERSION        :String[6]='0.0.62';
 BUFSIZE        = 16384;

 {$I BALDUR.INC}

VAR
 F              :File;
 BytesRead      :Word;

 S              :String;
 i              :LongInt;
 w1             :Word;

 Buf            :Array[1..BufSize] of byte;

 Extracted      :LongInt;

 RESoutdir,
 KEYfn          :String[128];
 BIFfn          :String[20];
 RESfn          :String[128]; { as found in key index, with path }

 isFound        :Boolean;
 doKEY          :Boolean;
 doDIR          :Boolean;
 doDecrypt      :Boolean;
 doExtr         :Boolean;
 doExtrType     :Word;


 KEYResIdx      :LongInt;
 KEYIndex       :LongInt;
 KEYHeader      :TKEYHeader;
 KEYResFile     :TKEYResFileEntry;
 KEYResource    :TKEYResourceEntry;

 BIFF           :File;
 BIFHeader      :TBIFHeader;
 BIFCatEntry    :TBIFCatEntry;

Procedure Error(CONST errmsg: String);
begin
 WriteLn(' ! ERROR: ',errmsg);
 Halt(1);
end;

Procedure Warning(CONST warnmsg: String);
begin
 WriteLn(' ? Warning: ',warnmsg);
end;

Function GetResIDExt(const ResID: Word): String;
var
 Ext    :String[3];
begin
 ext:='';
 case ResID of
  $0001: ext:='BMP';
  $0002: ext:='MVE';
  $0004: ext:='WAV';
  $03E8: ext:='BAM';
  $03E9: ext:='WED';
  $03EA: ext:='CHU';
  $03EB: ext:='';
  $03EC: ext:='MOS';
  $03ED: ext:='ITM';
  $03EE: ext:='SPL';
  $03EF: ext:='BCS';
  $03F0: ext:='LUA';
  $03F1: ext:='CRE';
  $03F2: ext:='ARE';
  $03F3: ext:='DLG';
  $03F4: ext:='2DA';
  $03F5: ext:='GAM';
  $03F6: ext:='STO';
  $03F7: ext:='WMA';
  $03F8: ext:='EFF';
 end;
 GetResIDExt:=ext;
end;

Procedure OpenBIF(const filename: String);
begin
 WriteLn('  Opening BIF V1-resource "'+BIFfn+'"');
 Assign(BIFF,filename);
 Reset(BIFF,1);
 BlockRead(BIFF,BIFHeader,SizeOf(BIFHeader),BytesRead);
end;

Procedure ExtractBIF(const ResIdx: Word);
var
 S,S2           :String[8+1+3];
 F2             :File;
 j,k            :LongInt;
 IsEncrypted    :Boolean;
 KeyPtr         :Word;

 Procedure doCheckDecryptBlock;
 var
  i     :Word;
 begin
  if (IsEncrypted) and (DoDecrypt) then
   for i:=1 to BytesRead do begin
    Buf[i]:=Buf[i] XOR XORTable[KeyPtr];
    KeyPtr:=(KeyPtr+1) mod SizeOf(XORTable);
   end;
 end;

begin
 Seek(BIFF,BIFHeader.CatOffset + (ResIdx*SizeOf(BIFCatEntry)));
 BlockRead(BIFF,BIFCatEntry,SizeOf(BIFCatEntry),BytesRead);
 if ResIdx<>BIFCatEntry.ResIdx then begin
  Error('ResIdx mismatch - must do O(n) search. Not Implemented. Aborting.');
 end;

 S[0]:=#8;
 Move(KEYResource.Identifier,S[1],8);
 While S[length(S)]=#0 do Dec(S[0]);

 IsEncrypted:=False;
 KeyPtr:=0;

 S2:=GetResIDExt(BIFCatEntry.ChunkType);

 if KEYResource.ResType=$4 { special ID of sound }
 then begin
  Seek(BIFF,BIFCatEntry.Offset);
  BlockRead(BIFF,S2[1],4,BytesRead);
  S2[0]:=#4;
  if S2='WAVC' then S2:='WAC' else S2:=GetResIDExt(BIFCatEntry.ChunkType);
 end else
 if (KEYResource.ResType=$03F4) or (KEYResource.ResType=$03F0)
 then begin { check if encrypted resource }
  Seek(BIFF,BIFCatEntry.Offset);
  BlockRead(BIFF,S2[1],2,BytesRead);
  if (S2[1]=Char($FF)) and (S2[2]=Char($FF)) then IsEncrypted:=True;
  S2:=GetResIDExt(BIFCatEntry.ChunkType);
 end;

 S:=S+'.'+S2;

 Write('  Extracting resource #',HexW(ResIdx),' "'+ResOutDir+'\'+S+'"');

 if IsEncrypted then begin
  if doDecrypt then begin
   Write(' [decrypted]');
   Dec(BIFCatEntry.Size,2);
   Inc(BIFCatEntry.Offset,2);
  end else
   Write(' [encrypted]');
 end;

 j:=BIFCatEntry.Size div BUFSIZE;
 Seek(BIFF,BIFCatEntry.Offset); { make sure we're at the right place }

 if NOT Exist(ResOutDir) then
 begin
  {$I-}
  MkDir(ResOutDir);
  {$I+}
  if IOResult<>0 then Error('Error creating directory "'+ResoutDir+'"');
 end;

 Assign(F2,ResOutDir+'\'+S);
 ReWrite(F2,1);
 for k:=1 to j do begin
  BlockRead(BIFF,Buf,SizeOf(Buf),BytesRead);
  doCheckDecryptBlock;
  BlockWrite(F2,Buf,BytesRead);
 end; { .. main blocks of file }

  k:=BIFCatEntry.Size mod BUFSIZE;
  BlockRead(BIFF,Buf,k,BytesRead);
  doCheckDecryptBlock;
  BlockWrite(F2,Buf,BytesRead);

  Close(F2);
  WriteLn;
  Inc(Extracted); { increase count }
end;

Procedure Info;
begin
 WriteLn(' Usage: BIFTOOL [switches] <filename>');
 WriteLn;
 WriteLn(' Switches:');
 WriteLn('  -extr [$ResId]         ; Extract resources.');
 WriteLn('  -nodecrypt             ; Do not decrypt encrypted resources.');
 WriteLn('  -dir                   ; Print directory. (Default behaviour)');
 WriteLn;
 WriteLn(' Notes:');
 WriteLn('  Set the envvar BG to your Baldur''s Gate installation path.');
 WriteLn('  Contact the authors thru srm_dfr@hotmail.com');
 Halt(0);
end;

BEGIN
 WriteLn;
 WriteLn('BIFTOOL v'+Version+' - (C)1999 Eddy & Robert / DFR Research & Engineering');
 WriteLn('srm_dfr@hotmail.com - http://hem2.passagen.se/eddy1/gamehack/baldur.html');
 WriteLn;

 if (ParamCount=0) or (ParamExist('/?')) or (ParamExist('-help')) then Info;

 doExtr:=ParamExist('-extr');
 doDIR:=ParamExist('-dir');
 doDecrypt:=NOT ParamExist('-nodecrypt');
 BIFfn:=ParamStr(ParamCount);

 S:=ParamValue('-extr'); Delete(S,1,1);
 DoExtrType:=Va(S);

 if doExtr then begin
  if NOT exist(BIFfn) then begin
   Error('File "'+BIFfn+'" not found. Extraction not possible.');
   doExtr:=False;
  end else begin
   Extracted:=0;
   Write('  Extraction enabled: ');
   if doExtrType=0 then WriteLn('All types.') else WriteLn('Type 0x',HexW(doExtrType),
    ' (=".',GetResIDExt(doExtrType),'") only.');
   OpenBIF(BIFfn);
   if BIFHeader.CatMMEntries>0 then Warning(St(BIFHeader.CatMMEntries)+ ' chunks of 0x3EB resources will be skipped.');
   ResOutDir:=FileBase(BIFfn);
  end;
 end;

 KEYfn:=GetEnv('BG');
 if (length(KEYfn)>0) and (KEYfn[length(KEYfn)]<>'\') then KEYfn:=KEYfn+'\';
 KEYfn:=KEYfn+'chitin.key';

 if NOT Exist(KEYfn) then begin
  KEYfn:='..\chitin.key';
 end;

 if NOT Exist(KEYfn) then begin
  Warning('Couldn''t locate KEY V1-resource "'+KEYfn+'"');
  doKEY:=False;
 end else begin
  WriteLn('  Opening KEY V1-resource "'+KEYfn+'"');
  doKEY:=True;
 end;

 if doExtr and doDir then
 begin
  Warning('-extr and -dir are mutually exclusive. Disabling -dir.');
  doDir:=False;
 end;

 if (NOT doExtr) and (NOT doDir) then
 begin
  Warning('No options specified. Defaulting to -dir.');
  doDir:=True;
 end;

 if doKEY then begin
  Assign(F,KEYfn);
  Reset(F,1);
  BlockRead(F,KEYHeader,SizeOf(KEYHeader),BytesRead);

  RESfn:='';
  isFound:=FALSE;
  for KeyIndex:=1 to KEYHeader.ResFileEntries do
  begin
   { move to the correct record }
   Seek(F,KEYHeader.FileTblOffs + ((KeyIndex-1)*SizeOf(KEYResFile)));
   BlockRead(F,KEYResFile,SizeOf(KEYResFile),BytesRead);

   S[0]:=Char(KEYResFile.FilenameLen-1);
   Seek(F,KEYResFile.FilenameOffs);
   BlockRead(F,S[1],Byte(S[0]),BytesRead);
   if Right(Upper_F(S),Length(BIFfn))=Upper_F(BIFfn) then
    begin
     RESfn:=S;
     Dec(KeyIndex);
     WriteLn('  Located "',RESfn,'" at index 0x',HexW(KeyIndex));
     isFound:=TRUE;
     Break;
    end;
   end;

  if isFound then begin
   WriteLn('  Directory of resources in "',RESfn,'" available.');
   Seek(F,KEYHeader.ResTblOffs);

   KeyResIdx:=0;
   for i:=1 to KEYHeader.ResEntries do
   begin
    BlockRead(F,KEYResource,SizeOf(KEYResource),BytesRead);
    w1:=Va('$'+Copy( HexW(KEYResource.Unknown2),2,2));

    if (w1=KeyIndex) then { Only the relevant files }
    begin

     if doDIR then begin
      S[0]:=#8;
      Move(KEYResource.Identifier,S[1],8);
      WriteLn(HexW(i),':',HexW(KeyResIdx),' ',S,' .',
       GetResIDExt(KEYResource.ResType),' (',HexW(KEYResource.ResType),')');
     end;

     if DoExtr AND (KEYResource.ResType=$03EB) then Warning('Skipping 0x3EB resource.') else
     begin
      if DoExtr AND ((DoExtrType=0) OR (DoExtrType=KEYResource.ResType)) then ExtractBif(KeyResIdx);
      Inc(KeyResIdx);
     end; { .. not MM res }

    end;
   end;
  end else begin
   Warning('.BIF not found in KEY index. No action taken.');
   doKey:=False;
  end;
  Close(F); { Close open KEY }
 end;

 if doExtr then begin
  Close(BIFF); { close open BIF }
  WriteLn('  ',Extracted,' resources were extracted.');
 end;

END.
